/*
 * Copyright (c) 2011 Sprite_tm <jeroen@spritesmods.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * For more information on the GPL, please go to:
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <avr/io.h>
#include <avr/pgmspace.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>

#include "config.h"
#include "core/eeprom.h"
#include "hardware/z80interface/z80interface.h"
#include "protocols/uip/uip.h"
#include "protocols/uip/parse.h"
#include "arcade1943.h"


//Map of tile<->ascii-character mappings. Hope this is somewhat correct...
static const char tilemap[] PROGMEM=
	"0123456789ABCDEF" //00
	"GHIJKLMNOPQRSTUV" //10
	"WXYZ      ,.:%*#" //20
	"()<>  +-= &    `" //30
	"      !!??      " //40
	"          &     " //50
	"                "
	"                "
	"         JKLMNOP" //60
	"QRSTUVWXYZabcdef" //70
	"ghijklmnopqrstuv"
	"wxyz1234567890&?"
	"()_.. ,'\"\" '\"-/%"
	"       *~       "
	"\000";

#define TILE_EMPTY 0x24
#define DELAY_BEFORE_START_SEC 20

/* At the location of HIGHSCORE_START, there are 5 16-byte highscore entries.
Format of a high score entry:
- 8 bytes of single-digit decimal code (00 00 00 09 01 05 00 00 = 91500)
- 3 bytes of tile-encoded initials
- 5 bytes of garbage (0x24) */
#define HIGHSCORE_START 0xE600

//Location of value of top highscore indicator
//Doesn't actually update much on screen, that has to be done separately :/
#define HIGHSCORE_TINY_START 0xE110 

//Location of the _end_ of the on-screen top hiscore.
#define HIGHSCORE_DISPLAY_END 0xD27E
#define HIGHSCORE_DISPLAY_STRIDE 32

#define SCREENLEN 28
#define MESSAGE_START 0xD063
#define MESSAGE_STRIDE 32

static int realInitDelaySec;
static char message[SCREENLEN];
static int messageTimerSec=0;
static arcade1943HighScoreStruct highScores;


//Convert an ascii character to a tile number.
static char asciiToTile(char c) {
	char *p=tilemap;
	while (pgm_read_byte(p)!=0 && pgm_read_byte(p)!=c) {
		p++;
	}
	if (pgm_read_byte(p)==0) return TILE_EMPTY; //unknown char
	return (char)(p-tilemap);
}

//Convert a tile number to an ascii character.
static char tileToAscii(char c) {
	return pgm_read_byte(&tilemap[c]);
}

//Write the hiscore from the highScores struct to the machines memory.
//Needs to update counters in various places in memory...
static void writeHiscoreToMachine(void) {
	int x,y;
	unsigned int addr;
	unsigned long score;
	//Sanity check
	if (highScores.sc[0].score<highScores.sc[1].score) return;
	if (highScores.sc[1].score<highScores.sc[2].score) return;
	if (highScores.sc[2].score<highScores.sc[3].score) return;
	if (highScores.sc[3].score<highScores.sc[4].score) return;

	//Need to write hiscore to three locations: hiscore table, hiscore indicator on display and hiscore
	//counter internal to machine
	//Write hiscore table
	for (x=0; x<5; x++) {
		z80TakeOverBus();
		score=highScores.sc[x].score;
		addr=HIGHSCORE_START+(x*16)+7;
		for (y=0; y<8; y++) {
			z80SetMem(addr, score%10);
			score=score/10;
			addr--;
		}
		addr=HIGHSCORE_START+(x*16)+8;
		z80SetMem(addr++, highScores.sc[x].initials[0]);
		z80SetMem(addr++, highScores.sc[x].initials[1]);
		z80SetMem(addr++, highScores.sc[x].initials[2]);
		z80ReleaseBus();
	}
	z80TakeOverBus();

	//Write internal hiscore ctr
	score=highScores.sc[0].score;
	addr=HIGHSCORE_TINY_START+7;
	for (x=0; x<8; x++) {
		z80SetMem(addr, score%10);
		score=score/10;
		addr--;
	}

	//Update display, make sure not to write initial zeroes
	score=highScores.sc[0].score;
	addr=HIGHSCORE_DISPLAY_END;
	for (x=0; x<8; x++) {
		if (score || x==0) {
			z80SetMem(addr, score%10);
		} else {
			z80SetMem(addr, TILE_EMPTY);
		}
		score=score/10;
		addr-=HIGHSCORE_DISPLAY_STRIDE;
	}

	//All done.
	z80ReleaseBus();
}

//Show a message on the machines screen for timeSec seconds.
//Actually just prepares it (ascii->tiles), the arcade1943_tick routine does
//the real work.
void arcade1943_setMessage(char* msg, int timeSec) {
	int p=0;
	for (p=0; p<SCREENLEN; p++) message[p]=TILE_EMPTY;
	p=0;
	while (p!=SCREENLEN && msg[p]!=0) {
		message[p]=asciiToTile(msg[p]);
		p++;
	}
	messageTimerSec=timeSec;
}

//Convert a hiscore position to a string
int arcade1943_getHiScoreString(char *str, int pos) {
	int l;
	if (pos>=5) return 0;
	l=sprintf(str, "%li ", highScores.sc[pos].score);
	str[l++]=tileToAscii(highScores.sc[pos].initials[0]);
	str[l++]=tileToAscii(highScores.sc[pos].initials[1]);
	str[l++]=tileToAscii(highScores.sc[pos].initials[2]);
	return l;
}

//First init
void arcade1943_init() {
	char msg[]="Requesting IP";
	realInitDelaySec=DELAY_BEFORE_START_SEC;
	//Load highscore from eeprom
	eeprom_restore(arcade1943HighScore, &highScores, sizeof(highScores));
	_delay_ms(100);
	writeHiscoreToMachine();
	arcade1943_setMessage(msg, 20);
}

//Called after we're sure both the arcade machine and the network are up and running.
void arcade1943_init_real() {
	char msg[16];
	//Fetch our IP and dump it on the machines screen.
    uip_ipaddr_t hostaddr;
	uip_gethostaddr(&hostaddr);
	print_ipaddr(&hostaddr, msg, 16);
	arcade1943_setMessage(msg, 10);
}

//Called once per second.
void arcade1943_tick() {
	int x,y;
	unsigned int a;
	long score;
	char initials[3];
	char weeprom;
	
	//Perhaps we have to call the initial init?
	if (realInitDelaySec>0) {
		realInitDelaySec--;
		if (realInitDelaySec==0) arcade1943_init_real();
	}
	//Grab highscores; commit to eeprom if changed.
	cli();
	z80TakeOverBus();
	weeprom=0;
	for (x=0; x<5; x++) {
		a=HIGHSCORE_START+(x*16);
		score=0;
		for (y=0; y<8; y++) {
			score=score*10;
			score=score+z80ReadMem(a++);
		}
		for (y=0; y<3; y++) initials[y]=z80ReadMem(a++);
		if (highScores.sc[x].score!=score) weeprom=1;
		if (highScores.sc[x].initials[0]!=initials[0]) weeprom=1;
		if (highScores.sc[x].initials[1]!=initials[1]) weeprom=1;
		if (highScores.sc[x].initials[2]!=initials[2]) weeprom=1;
		highScores.sc[x].score=score;
		highScores.sc[x].initials[0]=initials[0];
		highScores.sc[x].initials[1]=initials[1];
		highScores.sc[x].initials[2]=initials[2];
	}

	//See if we have to show a message.
	if (messageTimerSec>0) {
		messageTimerSec--;
		if (messageTimerSec==0) {
			//Message timed out. Blank message.
			for (x=0; x<SCREENLEN; x++) message[x]=TILE_EMPTY;
		}
		//Write to vram.
		a=MESSAGE_START;
		for (x=0; x<SCREENLEN; x++) {
			z80SetMem(a, message[x]);
			a+=MESSAGE_STRIDE;
		}
	}
	z80ReleaseBus();
	
	//Check if machine reset under us; in that case we need to re-write the
	//hiscores back to the machines ram again.
	if (highScores.sc[4].score==50000 || realInitDelaySec!=0) {
		eeprom_restore(arcade1943HighScore, &highScores, sizeof(highScores));
		writeHiscoreToMachine();
		weeprom=0;
	}
	sei();

	//need to update internal hs table?
	if (weeprom && realInitDelaySec==0) {
		eeprom_save(arcade1943HighScore, &highScores, sizeof(highScores));
		eeprom_update_chksum();
	}
}


/*
  -- Ethersex META --
  init(z80init)
  init(arcade1943_init)
  timer(50, arcade1943_tick())
*/


